__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)

    uri = None
    cache_dir = None

    # If we have one argument, if can be either the uri or cache
    if len(args.paths) >= 1:
        if args.paths[0].startswith("ghcr.io"):
            uri = args.paths.pop(0)
        else:
            cache_dir = args.paths.pop(0)

    # Do we still have the other one?
    if args.paths and not uri:
        uri = args.paths.pop(0)
    elif args.paths and not cache_dir:
        cache_dir = args.paths.pop(0)

    # Do we want to push to a build cache?
    cache = cli.push(cache_dir=cache_dir, uri=uri, tag=args.tag)

    # For an explicit push, the cache dir needed to already exist, require --cleanup
    if args.cleanup:
        cache.remove()



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

__version__ = "0.0.14"
AUTHOR = "Vanessa Sochat, Alec Scott"
NAME = "pakages"
PACKAGE_URL = "https://github.com/syspack/pakages"
KEYWORDS = "software, GitHub packages."
DESCRIPTION = "GitHub packages package manager."
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("pyaml", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("requests", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.repo import PakRepo

import re
import os


def parse_package_request(args):
    """
    Given args, parse the packages and repo (if any) from it.
    This is shared between install and build
    """
    # By defualt, assume not adding a repository
    repo = None

    # Case 1: we have an install directed at the present working directory
    if args.packages and args.packages[0] == ".":
        repo = os.getcwd()
        args.packages.pop(0)

    # If we have a path (akin to the first)
    if args.packages and os.path.exists(args.packages[0]):
        repo = args.packages.pop(0)

    # OR if we have a github URI
    if args.packages and re.search("(http|https)://github.com", args.packages[0]):
        repo = args.packages.pop(0)

    # Add the repository, if defined
    if repo:
        repo = PakRepo(repo)

    # If we don't have packages and we have a repo, derive from PWD
    if repo and not args.packages:
        args.packages = repo.list_packages()
    return args.packages, repo



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import pakages.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# The default settings file in the install root
default_settings_file = os.path.join(reps["$install_dir"], "settings.yml")

# Pakages home stores user specific keys, etc
pakageshome = os.path.expanduser("~/.pakages")

# The user settings file can be created to over-ride default
user_settings_file = os.path.join(pakageshome, "settings.yml")

# Default directory for keys
keys_dir = os.path.join(pakageshome, "keys")

# Fallback content type
content_type = "application/vnd.spack.package"

# Allowed variables for the environment
allowed_envars = ["username"]

# Default registry prefix to install from
trusted_packages_registry = "ghcr.io/pakages"

# The GitHub repository for pakages
github_url = "https://github.com/syspack/pakages"



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)
    cli.uninstall(args.packages)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient


def main(args, parser, extra, subparser):

    lookup = {"ipython": ipython, "python": python}
    shells = ["ipython", "python"]

    # The default shell determined by the command line client
    shell = args.interpreter
    if shell is not None:
        shell = shell.lower()
        if shell in lookup:
            try:
                return lookup[shell](args)
            except ImportError:
                pass

    # Otherwise present order of liklihood to have on system
    for shell in shells:
        try:
            return lookup[shell](args)
        except ImportError:
            pass


def create_client(args):
    return PakClient(settings_file=args.settings_file)


def ipython(args):
    """
    Generate an IPython shell with the client.
    """
    client = create_client(args)
    from IPython import embed

    embed()


def python(args):
    """
    Generate an python shell with the client.
    """
    import code

    client = create_client(args)
    code.interact(local={"client": client})



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import shutil
import tarfile
from collections import defaultdict

import pakages.oras
import pakages.sbom
import pakages.utils
import pakages.defaults
from pakages.logger import logger

import spack.binary_distribution as bd
import spack.hooks
import spack.store
import spack.util.crypto


def do_install(self, **kwargs):
    """
    Refactored install process to add proper cache handling.
    """
    from spack.installer import PackageInstaller

    dev_path_var = self.spec.variants.get("dev_path", None)
    registries = kwargs.get("registries")
    tag = kwargs.get("tag")

    if dev_path_var:
        kwargs["keep_stage"] = True

    class PakInstaller(PackageInstaller):
        """
        The PakInstaller wraps the PackageInstaller.

        We take this inline approach because we are not able to import it.
        """

        def _get_task(self, pkg, request, is_compiler, all_deps):
            """
            Get a build task for the package.
            """
            task = spack.installer.BuildTask(
                pkg,
                request,
                is_compiler,
                0,
                0,
                spack.installer.STATUS_ADDED,
                self.installed,
            )
            pkg_id = spack.installer.package_id(pkg)

            for dep_id in task.dependencies:
                all_deps[dep_id].add(pkg_id)
            return task

        def prepopulate_tasks(self):
            """
            Pre-populate a lookup of tasks for the cache.
            """
            # Prepare the complete list of dependencies to look for
            # This mimicks init_queue
            self._pakages_tasks = {}
            all_deps = defaultdict(set)
            logger.debug("Initializing the build queue from the build requests")
            for request in self.build_requests:
                for dep in request.traverse_dependencies():
                    dep_id = spack.installer.package_id(dep.package)
                    if dep_id not in self.build_tasks:
                        task = self._get_task(dep.package, request, False, all_deps)
                        self._pakages_tasks[task.pkg_id] = task
                    spack.store.db.clear_failure(dep, force=False)

                if request.pkg_id not in self._pakages_tasks:
                    spack.store.db.clear_failure(request.spec, force=True)
                    task = self._get_task(request.pkg, request, False, all_deps)
                    self._pakages_tasks[task.pkg_id] = task

            # Add missing dependents to ensure proper uninstalled dependency
            # tracking when installing multiple specs
            for dep_id, dependents in all_deps.items():
                if dep_id in self._pakages_tasks:
                    task = self._pakages_tasks[dep_id]
                    for dependent_id in dependents.difference(task.dependents):
                        task.add_dependent(dependent_id)

        def prepare_cache(self, registries=None, tag=None):
            """
            Given that we have a build cache for a package, install it.

            Since the GitHub packages API requires a token, we take an approach
            that attempts a pull for an artifact, and just continue if we don't
            have one.
            """
            # If no registries in user settings or command line, use default
            if not registries:
                registries = [pakages.defaults.trusted_packages_registry]
            tag = tag or pakages.defaults.default_tag

            # prepare oras client
            oras = pakages.oras.Oras()

            if not self.build_requests:
                return

            # If we want to use Github packages API, it requires token with package;read scope
            # https://docs.github.com/en/rest/reference/packages#list-packages-for-an-organization
            for pkg_id, request in self._pakages_tasks.items():

                # Don't continue if installed!
                if request.pkg.spec.install_status() == True:
                    continue

                # The name of the expected package, and directory to put it
                tmpdir = pakages.utils.get_tmpdir()

                # Try until we get a cache hit
                artifact = None
                for registry in registries:
                    name = bd.tarball_name(request.pkg.spec, ".spack")
                    generic_name = pakages.utils.generalize_spack_archive(name)
                    uri = "%s/%s:%s" % (registry, generic_name, tag)
                    artifact = oras.fetch(uri, os.path.join(tmpdir, name))
                    if artifact:
                        break

                # Don't continue if not found
                if not artifact:
                    shutil.rmtree(tmpdir)
                    continue

                # Checksum check (removes sha256 prefix)
                sha256 = oras.get_manifest_digest(uri)
                if sha256:
                    checker = spack.util.crypto.Checker(sha256)
                    if not checker.check(artifact):
                        logger.exit("Checksum of %s is not correct." % artifact)

                # If we have an artifact, extract where needed and tell spack it's installed!
                if artifact:
                    logger.info(f"Extracting archive {artifact}...")

                    # Note - for now not signing, since we don't have a consistent key strategy
                    # The spack function is a hairball that doesn't respect the provided filename
                    spec = extract_tarball(request.pkg.spec, artifact)
                    if not spec:
                        continue
                    # Update the build request
                    #                    request.pkg.spec = spec
                    #                   request.pkg_id = spack.installer.package_id(spec.package)
                    #                  request.dependencies = {f"{x.name}-{x.version}-{x._hash}" for x in spec.dependencies()}
                    #                 self.build_requests[i] = request

                    # And finish this piece of the install
                    spack.hooks.post_install(spec)
                    spack.store.db.add(spec, spack.store.layout)

    builder = PakInstaller([(self, kwargs)])

    # Download what we can find from the GitHub cache
    builder.prepopulate_tasks()
    builder.prepare_cache(registries, tag)
    builder.install()

    # If successful, generate an sbom
    meta_dir = os.path.join(self.prefix, ".spack")
    if os.path.exists(meta_dir):
        sbom = pakages.sbom.generate_sbom(self.spec)
        sbom_file = os.path.join(meta_dir, "sbom.json")
        pakages.utils.write_json(sbom, sbom_file)


def extract_tarball(spec, filename):
    """
    extract binary tarball for given package into install area
    """
    from_dir = os.path.dirname(filename)
    name = os.path.basename(filename)
    dag_hash = name.split("-")[-1].replace(".spack", "")

    # Anything installed from system
    if spack.store.layout.root not in spec.prefix:
        extract_dir = spack.store.layout.root

        # ['linux', 'ubuntu20.04', 'x86_64', 'gcc', '9.4.0', 'bzip2', '1.0.8']
        parts = os.path.basename(filename).split("-")[:-1]
        prefix = os.path.join(
            extract_dir,
            f"{parts[0]}-{parts[1]}-{parts[2]}",
            f"{parts[3]}-{parts[4]}",
            f"{spec.name}-{spec.version}-{dag_hash}",
        )
    else:
        extract_dir = os.path.dirname(spec.prefix)
        prefix = os.path.join(extract_dir, f"{spec.name}-{spec.version}-{dag_hash}")

    # The spec prefix is an easy way to get the directory name
    if os.path.exists(prefix):
        shutil.rmtree(prefix)

    tmpdir = pakages.utils.get_tmpdir()
    with tarfile.open(filename, "r") as tar:
        tar.extractall(tmpdir)

    # Find the json file and .tar.gz
    spec_file = None
    targz = None
    for path in os.listdir(tmpdir):
        if path.endswith(".json"):
            spec_file = os.path.join(tmpdir, path)
        elif path.endswith("gz"):
            targz = os.path.join(tmpdir, path)

    if not spec_file:
        shutil.rmtree(tmpdir)
        shutil.rmtree(from_dir)
        logger.exit(f"{spec} did not come with a spec json.")

    new_prefix = str(os.path.relpath(prefix, spack.store.layout.root))
    content = pakages.utils.read_json(spec_file)

    # if the original relative prefix is in the spec file use it
    buildinfo = content.get("buildinfo", {})
    relative_prefix = buildinfo.get("relative_prefix", new_prefix)

    extract_tmp = os.path.join(spack.store.layout.root, ".tmp")
    pakages.utils.mkdirp(extract_tmp)
    extracted_dir = os.path.join(extract_tmp, relative_prefix.split(os.path.sep)[-1])

    with tarfile.open(targz, "r") as tar:
        tar.extractall(path=extract_tmp)
    try:
        shutil.move(extracted_dir, prefix)
    except Exception as e:
        shutil.rmtree(extracted_dir)
        raise e

    # Clean up
    os.remove(targz)
    os.remove(spec_file)

    # Create a dummy spec to do the relocation
    new_spec = spack.spec.Spec.from_dict({"spec": content["spec"]})

    try:
        bd.relocate_package(new_spec, False)
    # This probably shouldn't fail, will let it slide for now
    except Exception as e:
        raise e
    finally:
        shutil.rmtree(tmpdir)
        if os.path.exists(filename):
            os.remove(filename)
    return new_spec



from .terminal import (
    run_command,
    check_install,
    get_installdir,
    get_username,
    which,
    confirm_action,
)
from .spack import (
    add_spack_to_path,
    add_pakages_spack_repo,
    install_spack,
    generalize_spack_archive,
)
from .fileio import (
    copyfile,
    get_file_hash,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    mkdirp,
    print_json,
    read_file,
    read_json,
    read_yaml,
    recursive_find,
    workdir,
    write_file,
    write_json,
)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

# Generate a software bill of materials for a spack package

# https://www.ntia.gov/files/ntia/publications/howto_guide_for_sbom_generation_v1.pdf

import spack.main
import spack.config
import spack.spec

from datetime import datetime
import uuid

# We will use CycloneDX which is a simplified format approved by standards committees
# https://cyclonedx.org/docs/1.3/json
# https://cyclonedx.org/use-cases/
template = {
    # Specifies the format of the BOM. This helps to identify the file as CycloneDX
    # since BOMs do not have a filename convention nor does JSON schema support namespaces.
    "bomFormat": "CycloneDX",
    # The version of the CycloneDX specification a BOM is written to (starting at version 1.2)
    "specVersion": "1.3",
    "serialNumber": "",
    # The version WE are generating for the package - this would increment
    "version": 1,
    # Will be populated with metadata
    "metadata": {},
}


def generate_timestamp_now():
    return str(datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ"))


def get_component(spec):
    """
    Given a spec, get a component for it.
    """
    # Let's take a simple approach - anything with "lib" is a library, otherwise application
    component_type = "lib" if "lib" in spec.package.name else "application"

    # Must be concrets
    spec.concretize()

    component = {
        # Note that "application" might also be a suitable choice
        # https://cyclonedx.org/docs/1.3/json/#metadata_component_type
        "type": component_type,
        # Specifies the scope of the component. If scope is not specified,
        #'required' scope should be assumed by the consumer of the BOM
        # Let's explicitly state that :)
        "scope": "required",
        # The name of the component. This will often be a shortened, single name of the component.
        # Examples: commons-lang3 and jquery
        "name": spec.package.name,
        # Associated mime-type. Spack doesn't have one officially, but this is what
        # I was using for oras push to an OCI registry
        "mime-type": "application/vnd.spack.package",
        # The grouping name or identifier. This will often be a shortened,
        # single name of the company or project that produced the component,
        # or the source package or domain name. Whitespace and special characters
        # should be avoided. Examples include: apache, org.apache.commons, and apache.org.
        "group": "spack.io",
        # The component version. The version should ideally comply with semantic versioning but is not enforced.
        "version": str(spec.version),
        # An optional identifier which can be used to reference the component elsewhere in the BOM. Every bom-ref should be unique.
        # We can use the build hash since it's unique to this spec
        "bom-ref": str(spec),
        # Excluded
        # publisher: The person(s) or organization(s) that published the component
        # author: The person(s) or organization(s) that authored the component
        # supplier: The organization that supplied the component. The supplier may often be the manufacturer, but may also be a distributor or repackager.
        # licenses: spack doesn't have enough metadata to make a call for sure on each
        # copyright: An optional copyright notice informing users of the underlying claims to copyright ownership in a published work.
        # purl: spack doesn't have these
        # swid: Specifies metadata and content for ISO-IEC 19770-2 Software Identification (SWID) Tags.
        # pedigree: Component pedigree is a way to document complex supply chain scenarios where
        # components are created, distributed, modified, redistributed, combined with other components, etc.
        # Pedigree supports viewing this complex chain from the beginning, the end, or anywhere in the middle.
        # It also provides a way to document variants where the exact relation may not be known.
        # components: I assume this means some kind of sub-component? Or component referenced within
        # the component elsewhere? For dependencies I'm going to describe them under dependencies.
        # evidence: Provides the ability to document evidence collected through various forms of extraction or analysis.
        # properites: extra name/value keypairs
    }

    # Specifies a description for the component
    # Only add if not empty!
    description = spec.package.format_doc()
    if description:
        component["description"] = description

    # Add hashes, whichever might exist
    # I looked for all unique hash types on 11/28
    # names = set()
    # for package in spack.repo.all_package_names():
    #    spec = spack.spec.Spec(package)
    #    for _, version in spec.package.versions.items():
    #        for alg, _ in version.items():
    #            names.add(alg)

    component["hashes"] = []
    for key, hashvalue in spec.package.versions.get(spec.version, {}).items():

        # Covers both sha256 and sha256sum
        if key.startswith("sha256"):
            component["hashes"].append({"alg": "SHA-256", "content": hashvalue})

    # External references?
    if spec.package.all_urls:
        component["externalReferences"] = []
        for url in spec.package.all_urls:
            component["externalReferences"].append({"type": "website", "url": url})

    # Finally, custom spack metadata (properties)
    component["properties"] = {
        "spack:dag_hash": spec.dag_hash(),
        "spack:spec": str(spec),
        "spack:build_spec": str(spec.build_spec),
        "spack:architecture": str(spec.architecture),
        "spack:variants": str(spec.variants),
        "spack:compiler": str(spec.compiler),
    }
    return component


def generate_sbom(pkg):
    """
    Generates the sbom based on best practices suggested in the guide.
    """
    bom = template.copy()
    if isinstance(pkg, spack.spec.Spec):
        spec = pkg
    else:
        spec = spack.spec.Spec(pkg)

    # Every BOM generated should have a unique serial number, even if the contents of the BOM
    # being generated have not changed over time. The process or tool responsible for
    # creating the BOM should create random UUID's for every BOM generated.
    bom["serialNumber"] = "urn:uuid:" + str(uuid.uuid4())

    # The date and time (timestamp) when the document was created.
    metadata = {"timestamp": generate_timestamp_now()}

    # The tool(s) used in the creation of the BOM.
    metadata["tools"] = [
        {
            "vendor": "Lawrence Livermore National Lab",
            "name": "Spack",
            "version": spack.main.get_version(),
        }
    ]

    # The person(s) who created the BOM. Authors are common in BOMs created through
    # manual processes. BOMs created through automated means may not have authors.
    metadata["authors"] = [
        {"name": "@vsoch", "email": "vsoch@users.noreply.github.com"}
    ]

    # The component that the BOM describes.
    metadata["component"] = get_component(spec)

    # Skipped:
    # manufacture: The organization that manufactured the component that the BOM describes.
    # supplier: The organization that supplied the component that the BOM describes.
    # The supplier may often be the manufacturer, but may also be a distributor or repackager.

    # Let's assume licenses under the bom is meant to describe spack, but not the component?
    metadata["licenses"] = [
        {"license": {"name": "MIT"}},
        {"license": {"name": "Apache-2.0"}},
    ]
    bom["metadata"] = metadata

    # We might also assume that properties on this level could be for spack
    # Could put spack properties here.
    deps = spec.dependencies()
    components = {}

    # Let's include all nested dependencies
    if deps:
        while deps:
            dep = deps.pop(0)
            deps = deps + dep.dependencies()

            if str(dep) not in components:
                components[str(dep)] = get_component(dep)

    if components:
        bom["components"] = components

    # Finally, add direct dependencies
    deps = spec.dependencies()
    if deps:
        bom["dependencies"] = []
        for dep in deps:
            # The bom-ref identifiers of the components that are dependencies of this dependency object.
            dependsOn = [str(x) for x in dep.dependencies()]
            bom["dependencies"].append({"ref": str(dep), "dependsOn": dependsOn})

    # Add an external reference for spack packages, GitHub, etc.
    bom["externalReferences"] = [
        {"type": "website", "url": "https://github.com/spack/spack"},
        {"type": "website", "url": "https://spack.github.io/packages"},
    ]
    return bom



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"


from pakages.logger import logger
import pakages.defaults as defaults
import pakages.utils
import pakages.schemas
import shutil

from datetime import datetime
import jsonschema
import os
import yaml


class EmptySettings:
    def __init__(self):
        """
        Create a new settings object not requiring a settings file.
        """
        # Set an updated time, in case it's written back to file
        self._settings = {"updated_at": datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")}
        self.settings_file = None

    def __str__(self):
        return "[pakages-settings]"

    def __repr__(self):
        return self.__str__()

    def validate(self):
        """
        Validate the loaded settings with jsonschema
        """
        jsonschema.validate(instance=self._settings, schema=pakages.schemas.settings)

    def inituser(self):
        """
        Create a user specific config in user's home.
        """
        user_home = os.path.dirname(defaults.user_settings_file)
        if not os.path.exists(user_home):
            os.makedirs(user_home)
        if os.path.exists(defaults.user_settings_file):
            logger.exit(
                "%s already exists! Remove first before re-creating."
                % defaults.user_settings_file
            )
        shutil.copyfile(self.settings_file, defaults.user_settings_file)
        logger.info("Created user settings file %s" % defaults.user_settings_file)

    def edit(self):
        """
        Interactively edit a config file.
        """
        if not self.settings_file or not os.path.exists(self.settings_file):
            logger.exit("Settings file not found.")

        # Make sure editor exists first!
        editor = pakages.utils.which(self.config_editor)
        if not os.path.exists(editor):
            logger.exit(
                "Editor '%s' not found! Update with pakages config set config_editor:<name>"
                % self.config_editor
            )
        os.system("%s %s" % (editor, self.settings_file))

    def get_settings_file(self, settings_file=None):
        """
        Get the preferred used settings file.
        """
        # Only consider user settings if the file exists!
        user_settings = None
        if os.path.exists(defaults.user_settings_file):
            user_settings = defaults.user_settings_file

        # First preference to command line, then user settings, then default
        return settings_file or user_settings or defaults.default_settings_file

    def load(self, settings_file=None):
        """
        Load the settings file into the settings object
        """
        # Get the preferred settings flie
        self.settings_file = self.get_settings_file(settings_file)

        # Exit quickly if the settings file does not exist
        if not os.path.exists(self.settings_file):
            logger.exit("%s does not exist." % self.settings_file)

        # Store the original settings for update as we go
        with open(self.settings_file, "r") as fd:
            self._settings = yaml.load(fd.read(), yaml.SafeLoader)

    def get(self, key, default=None):
        """
        Get a settings value, doing appropriate substitution and expansion.
        """
        value = self._settings.get(key, default)
        value = self._substitutions(value)
        # If we allow environment substitution, do it
        if key in defaults.allowed_envars and value:
            if isinstance(value, list):
                value = [os.path.expandvars(v) for v in value]
            else:
                value = os.path.expandvars(value)
        return value

    def __getattr__(self, key):
        """
        A direct get of an attribute, but default to None if doesn't exist
        """
        return self.get(key)

    def add(self, key, value):
        """
        Add a value to a list parameter
        """
        # We can only add to lists
        current = self._settings.get(key)
        if current and not isinstance(current, list):
            logger.exit("You cannot only add to a list variable.")

        if value not in current:
            # Add to the beginning of the list
            current = [value] + current
            self.change_validate(key, current)
            self._settings[key] = current
            logger.warning(
                "Warning: Check with pakages config edit - ordering of list can change."
            )

    def remove(self, key, value):
        """
        Remove a value from a list parameter
        """
        current = self._settings.get(key)
        if current and not isinstance(current, list):
            logger.exit("You cannot only remove from a list variable.")
        if not current or value not in current:
            logger.exit("%s is not in %s" % (value, key))
        current.pop(current.index(value))
        self._settings[key] = current
        self.change_validate(key, current)
        logger.warning(
            "Warning: Check with pakages config edit - ordering of list can change."
        )

    def set(self, key, value):
        """
        Set a setting based on key and value. If the key has :, it's nested
        """
        value = True if value == "true" else value
        value = False if value == "false" else value

        # List values not allowed for set
        current = self._settings.get(key)
        if current and isinstance(current, list):
            logger.exit("You cannot use 'set' for a list. Use add/remove instead.")

        # This is a reference to a dictionary (object) setting
        if ":" in key:
            key, subkey = key.split(":")
            self._settings[key][subkey] = value
        else:
            self._settings[key] = value

        # Validate and catch error message cleanly
        self.change_validate(key, value)

    def change_validate(self, key, value):
        """
        A courtesy function to validate a new config addition.
        """
        # Don't allow the user to add a setting not known
        try:
            self.validate()
        except jsonschema.exceptions.ValidationError as error:
            logger.exit(
                "%s:%s cannot be added to config: %s" % (key, value, error.message)
            )

    def _substitutions(self, value):
        """
        Given a value, make substitutions
        """
        if isinstance(value, bool) or not value:
            return value

        # Currently dicts only support boolean or null so we return as is
        elif isinstance(value, dict):
            return value

        for rep, repvalue in defaults.reps.items():
            if isinstance(value, list):
                value = [x.replace(rep, repvalue) for x in value]
            else:
                value = value.replace(rep, repvalue)

        return value

    def delete(self, key):
        if key in self._settings:
            del self._settings[key]

    def save(self, filename=None):
        """
        Save settings, but do not change order of anything.
        """
        filename = filename or self.settings_file
        if not filename:
            logger.exit("A filename is required to save to.")

        with open(filename, "w") as fd:
            yaml.dump(self._settings, fd)

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value


class Settings(EmptySettings):
    """
    The settings class is a wrapper for easily parsing a settings.yml file.

    We parse into a query-able class. It also gives us control to update settings,
    meaning we change the values and then write them to file. It's basically
    a dictionary-like class with extra functions.
    """

    def __init__(self, settings_file, validate=True):
        """
        Create a new settings object, which requires a settings file to load
        """
        self.load(settings_file)
        if validate:
            self.validate()

        # Set an updated time, in case it's written back to file
        self._settings["updated_at"] = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient
from pakages.logger import logger
from .helpers import parse_package_request


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)

    # Get list of packages and (optionally) repository
    packages, repo = parse_package_request(args)

    # Finally, add the repository and install packages
    if repo:
        cli.add_repository(repo.repo_dir)
    cache = cli.build(
        args.packages,
        cache_dir=args.cache_dir,
        key=args.key,
        registry=args.registry,
        tag=args.tag,
    )

    # We cannot have both
    if args.push and args.push_trusted:
        logger.exit("Please use only --push or --pushd")

    # Do we want to push to a build cache?
    if args.push or args.push_trusted:
        cache.push(
            args.push,
        )

        # By default, we clean up the build cache, unless a custom cache is used
        if not args.no_cleanup and not args.cache_dir:
            cache.remove()

        # Otherwse, require --no-cleanup and --force
        elif not args.no_cleanup and args.cache_dir and args.force:
            cache.remove()



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import spack.bootstrap
import spack.spec
import pakages.utils as utils
import pakages.defaults
import spack.util.executable
import spack.util.crypto

from pakages.logger import logger

import os
import random
import requests
import time


class Oras:
    def __init__(self):
        self._oras = None

    @property
    def oras(self):
        """
        Get the oras executable (easier to install on your computer over bootstrap)
        """
        if not self._oras:
            with spack.bootstrap.ensure_bootstrap_configuration():
                spec = spack.spec.Spec("oras")
                spack.bootstrap.ensure_executables_in_path_or_raise(
                    ["oras"], abstract_spec=spec
                )
                self._oras = spack.util.executable.which("oras")
        return self._oras

    def get_manifest(self, uri):
        """
        Use crane to get the image manifest
        """
        response = requests.get("https://crane.ggcr.dev/manifest/" + uri)
        if response.status_code == 200:
            return response.json()

    def get_manifest_digest(self, uri):
        """
        Get the first layer digest (the spack package archive)
        """
        response = self.get_manifest(uri)
        if not response:
            return

        layers = response.get("layers")
        if layers:
            digest = layers[0].get("digest")
            if digest:
                return digest.replace("sha256:", "")

    def push(self, uri, push_file, content_type=None, retry=5, sleep=1):
        """
        Push an oras artifact to an OCI registry
        """
        tries = 0
        content_type = content_type or pakages.defaults.content_type
        logger.info("Pushing oras {0}".format(uri))
        with utils.workdir(os.path.dirname(push_file)):
            while tries < retry:
                try:
                    return self._push(uri, push_file, content_type)
                except:
                    time.sleep(sleep)
                    sleep = sleep * 2**tries + random.uniform(0, 1)
                    tries += 1

    def _push(self, uri, push_file, content_type):
        """
        Helper to push that provides consistent metadata
        """
        self.oras(
            "push",
            uri,
            "--manifest-config",
            "/dev/null:application/vnd.unknown.config.v1+json",
            # GitHub does not honor this content type - it will return an empty artifact
            os.path.basename(push_file) + ":" + content_type,
        )

    def fetch(self, url, save_file):
        """
        Fetch an oras artifact from an OCI registry
        """
        # We don't have programmatic access to list, so we just try to pull
        logger.info("Fetching oras {0}".format(url))
        save_dir = os.path.dirname(save_file)
        try:
            self.oras("pull", url, "-a", "--output", save_dir)
        except:
            logger.info("{0} is not available for pull.".format(url))
            return

        # Files are technically saved with the hash, we just hope spack will use
        files = os.listdir(save_dir)
        if len(files) > 0:
            save_file = os.path.join(save_dir, files[0])

        # Return the file if exists
        if os.path.exists(save_file):
            return save_file



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import pakages.cache
from .settings import Settings

import spack.cmd
import spack.target
import spack.main
import spack.config

import json

import pakages.spec


class PakClient:
    """
    Pakages has a main controller for interacting with pakages.
    """

    def __init__(self, *args, **kwargs):
        settings_file = kwargs.get("settings_file")
        validate = kwargs.get("validate", True)
        if not hasattr(self, "settings"):
            self.settings = Settings(settings_file, validate=validate)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[pakages-client]"

    def iter_specs(self, packages, concretize=True):
        """
        A shared function to retrieve iterable of specs from packages
        """
        for spec in pakages.spec.parse_specs(packages):
            yield spec

    def list_installed(self):
        """
        List installed packages
        """
        find = spack.main.SpackCommand("find")
        print(find())
        return json.loads(find("--json"))

    def build(self, packages, cache_dir=None, key=None, registry=None, tag=None):
        """
        Build a package into a cache
        """
        # Prepare a cache directory
        cache = pakages.cache.BuildCache(
            cache_dir=cache_dir or self.settings.cache_dir,
            username=self.settings.username,
            email=self.settings.email,
            settings=self.settings,
        )

        # Install all packages, and also generate sboms
        specs = self.install(packages, registry=registry, tag=tag)

        # TODO how can we attach the sbom to the package (aside from being in archive?)
        cache.create(specs, key=key)
        return cache

    def push(self, uri, cache_dir=None, tag=None):
        """
        Given an existing cache directory, push known specs to a specific uri
        """
        # Prepare a cache directory
        cache = pakages.cache.BuildCache(
            cache_dir=cache_dir or self.settings.cache_dir,
            username=self.settings.username,
            email=self.settings.email,
            settings=self.settings,
        )
        cache.push(uri, tag=tag)
        return cache

    def add_repository(self, path):
        """
        Add a repository.

        Given a path that exists, add the repository to the
        underlying spack. If you need to add a GitHub uri, create a
        pakages.repo.PakRepo first.
        """
        repos = spack.config.get("repos")
        repos.insert(0, path)
        spack.config.set("repos", repos)

    def install(self, packages, registry=None, tag=None):
        """
        Install one or more packages.

        This eventually needs to take into account using the GitHub packages bulid cache
        """
        # Default to registries defined in settings
        registries = self.settings.trusted_pull_registries

        # Do we have an additional trusted registry provided on the command line?
        if registry:
            registries = [registry] + registries

        specs = []
        for spec in self.iter_specs(packages):
            logger.info("Preparing to install %s" % spec.name)

            # TODO here we actually want to insert preparing GitHub packages build cache
            # We should match based on a basic set of architectures we know the containers support
            # E.g., check the platform and spit an error if it's some niche HPC one.

            spec.package.do_install(
                force=True,
                registries=registries,
                tag=tag or self.settings.default_tag,
            )
            specs.append(spec)
        return specs

    def uninstall(self, packages):
        """
        Uninstall a spack package
        """
        for spec in self.iter_specs(packages):
            try:
                spec.package.do_uninstall(force=True)
            except Exception as ex:
                logger.exit(ex)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import pakages.install
import pakages.handlers.github
import pakages.utils as utils
from pakages.logger import logger

import spack.repo
import spack.config
import llnl.util.lang
import spack.util.naming as nm

import six
import shutil
import sys
import os
import re


class PakRepo:
    @property
    def is_remote(self):
        if re.search("(http|https)://github.com", self.raw):
            return True
        return False

    def __init__(self, path, add=True):
        """
        Create a new spack repo and add for spack to see.

        A Pak Repo can be a GitHub URL or a local path. It has a repos.yaml
        and packages/ directory.
        """
        self.raw = path
        if os.path.exists(path):
            self.path = os.path.abspath(path)
        elif self.is_remote:
            self.path = pakages.handlers.github.GitHub().clone(path)
        self.validate()

    def validate(self):
        """
        Validate that we have a repo, meaning we expect a repo.yml|yaml and packages
        """
        repo_yaml = list(utils.recursive_find(self.path, "repo[.](yml|yaml)"))
        if not repo_yaml:
            logger.exit("Cannot find repo.yaml or repo.yml anywhere in %s" % self.path)
        repo_yaml = repo_yaml[0]

        # Add directory with repo.yaml to spack just to list
        self.repo_dir = os.path.dirname(repo_yaml)

    def list_packages(self):
        """
        Given a filesystem repository, list packages there.

        We don't add the repository to be known by spack here, as we would want
        to do this when we install.
        """
        self.add()
        repo = Repo(self.repo_dir)
        return repo.all_package_names()

    def cleanup(self):
        """
        Cleanup a cloned repository - be careful don't run this for a local path!
        """
        if os.path.exists(self.repo_dir) and self.is_remote:
            shutil.rmtree(self.repo_dir)
            repos = spack.config.get("repos")
            updated = [r for r in repos if r != self.repo_dir]
            spack.config.set("repos", updated)

    def add(self):
        """
        Add the repository to be known to spack
        """
        repos = spack.config.get("repos")
        repos.insert(0, self.repo_dir)
        spack.config.set("repos", repos)


class RepoPath(spack.repo.RepoPath):
    def __init__(self, *repos):
        """
        Since we want control over customizing the package class, we return
        a pakages.repo.Repo instead.
        """
        self.repos = []
        self.by_namespace = nm.NamespaceTrie()

        self._provider_index = None
        self._patch_index = None
        self._tag_index = None

        # Add each repo to this path.
        for repo in repos:
            try:
                if isinstance(repo, six.string_types):
                    repo = Repo(repo)
                self.put_last(repo)
            except spack.repo.RepoError as e:
                logger.exit("Failed to initialize repository: '%s': %s" % (repo, e))

    def find_module(self, fullname, python_path=None):
        # Compatibility method to support Python 2.7
        return None
        # Note, might need to update this to be akin to parent


class Repo(spack.repo.Repo):
    @spack.repo.autospec
    def get(self, spec):
        """Returns the package associated with the supplied spec

        However we add custom functions provided by Pakages.
        """
        if spec.name is None:
            raise spack.repo.UnknownPackageError(None, self)

        if spec.namespace and spec.namespace != self.namespace:
            raise spack.repo.UnknownPackageError(spec.name, self.namespace)

        package_class = self.get_pkg_class(spec.name)

        # Monkey patch the class with the new install!
        package_class.do_install = pakages.install.do_install

        try:
            return package_class(spec)
        except spack.error.SpackError:
            # pass these through as their error messages will be fine.
            raise
        except Exception as e:
            logger.debug(e)

            # Make sure other errors in constructors hit the error
            # handler by wrapping them
            if spack.config.get("config:debug"):
                sys.excepthook(*sys.exc_info())
            raise spack.repo.FailedConstructorError(spec.fullname, *sys.exc_info())


def _singleton_path(repo_dirs=None):
    """
    Get a singleton repository path, along with pakages custom
    """
    repo_dirs = repo_dirs or spack.config.get("repos")
    if not repo_dirs:
        raise spack.repo.NoRepoConfiguredError(
            "Spack configuration contains no package repositories."
        )

    path = RepoPath(*repo_dirs)
    sys.meta_path.append(path)
    return path


#: Singleton repo path instance
path = llnl.util.lang.Singleton(_singleton_path)


def get(spec):
    """Convenience wrapper around ``spack.repo.get()``."""
    # TODO add support here for a spec name that is from GitHub
    # And then add to singleton path (do dynamically as function)
    return path.get(spec)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
import pakages.utils as utils
from spack.main import SpackCommand
import pakages.defaults
import pakages.settings
import pakages.oras

import shutil
import os

gpg = SpackCommand("gpg")
bc = SpackCommand("buildcache")


def gpg_init(dirname):
    """
    Init the gpg directory, setting permissions etc.
    """
    utils.mkdirp(dirname)
    os.chmod(dirname, 0o700)


def get_gpg_key():
    """
    Find the first key created for spack

    This would really be nicer if we didn't have to use gpg :)
    """
    key = None
    lines = gpg("list").split("\n")
    for i, line in enumerate(lines):
        if "Spack" in line:
            key = lines[i - 1].strip()
            break
    return key


class BuildCache:
    """
    A controller that makes it easy to create a build cache and install to it.
    """

    def __init__(self, cache_dir=None, username=None, email=None, settings=None):
        if not cache_dir:
            cache_dir = utils.get_tmpdir()
        self.cache_dir = cache_dir

        # Inherit settings from the client, or set to empty settings
        if not settings:
            settings = pakages.settings.EmptySettings()
        self.settings = settings

        # Use defautls for username and email if not provided
        # TODO eventually we want to store keys elsewhere
        username = username or utils.get_username()
        email = email or "%s@users.noreply.spack.io" % username

        # TODO how do we check if this is already created?
        gpg("init")
        gpg("create", username, email)

    def remove(self):
        """
        Delete the entire build cache
        """
        if self.cache_dir and os.path.exists(self.cache_dir):
            logger.warning("Removing %s" % self.cache_dir)
            shutil.rmtree(self.cache_dir)

    def create(self, specs, key=None):
        """
        Create the build cache with some number of specs

        Ideally we could do spack buildcache add but that isn't supported.
        """
        # If a key isn't defined, choose the first spack one we find
        if not key:
            key = get_gpg_key()
        format_string = " ".join(
            ["%s@%s/%s" % (s.name, s.version, s.dag_hash()) for s in specs]
        )

        command = ["create", "-r", "-a"]
        if key:
            command += ["-k", key]
        command += ["-d", self.cache_dir, format_string]
        print(bc(*command))

    def push(self, uri=None, tag=None):
        """
        Push the build cache to an OCI registry (compatible with ORAS)
        """
        tag = tag or self.settings.default_tag
        content_type = self.settings.content_type or pakages.defaults.content_type
        uri = uri or self.settings.trusted_packages_registry

        # Create an oras client
        oras = pakages.oras.Oras()

        # Find all .spack archives in the cache
        for archive in utils.recursive_find(self.cache_dir, ".spack"):

            # Only push generic name - the full hash never hits
            package_name = os.path.basename(archive)
            generic_name = utils.generalize_spack_archive(package_name)
            full_name = "%s/%s:%s" % (uri, generic_name, tag)
            oras.push(full_name, archive, content_type=content_type)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[pakages-build-cache]"



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import pakages.defaults as defaults
from pakages.client import PakClient
from pakages.logger import logger
import sys


def main(args, parser, extra, subparser):

    # If nothing provided, show help
    if not args.params:
        print(subparser.format_help())
        sys.exit(0)

    # The first "param" is either set of get
    command = args.params.pop(0)

    # If the user wants the central config file
    if args.central:
        args.settings_file = defaults.default_settings_file

    validate = True if not command == "edit" else False
    cli = PakClient(settings_file=args.settings_file, validate=validate)

    # For each new setting, update and save!
    if command == "init":
        return cli.settings.inituser()
    if command == "edit":
        return cli.settings.edit()
    elif command in ["set", "add", "remove"]:
        for param in args.params:
            if ":" not in param:
                logger.warning(
                    "Param %s is missing a :, should be key:value pair. Skipping."
                    % param
                )
                continue
            key, value = param.split(":", 1)
            if command == "set":
                cli.settings.set(key, value)
                logger.info("Updated %s to be %s" % (key, value))
            elif command == "add":
                cli.settings.add(key, value)
                logger.info("Added %s to %s" % (key, value))
            elif command == "remove":
                cli.settings.remove(key, value)
                logger.info("Removed %s from %s" % (key, value))

        # Save settings
        cli.settings.save()

    # For each get request, print the param pair
    elif command == "get":
        for key in args.params:
            value = cli.settings.get(key)
            value = value or "is unset"
            logger.info("%s %s" % (key.ljust(30), value))

    else:
        logger.error("%s is not a recognized command." % command)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.logger import logger
from subprocess import Popen, PIPE, STDOUT
import os
import shlex
import shutil
import getpass


def which(software, strip_newline=True):
    """get_install will return the path to where Singularity (or another
    executable) is installed.
    """
    return shutil.which(software)


def check_install(software, quiet=True, command="--version"):
    """check_install will attempt to run the singularity command, and
    return True if installed. The command line utils will not run
    without this check.

    Parameters
    ==========
    software: the software to check if installed
    quiet: should we be quiet? (default True)
    command: the command to use to check (defaults to --version)
    """
    cmd = [software, command]
    try:
        version = run_command(cmd, software)
    except:  # FileNotFoundError
        return False
    if version:
        if not quiet and version["return_code"] == 0:
            version = version["message"]
            logger.info("Found %s version %s" % (software.upper(), version))
        return True
    return False


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def get_username():
    return getpass.getuser()


def run_command(cmd, stream=False):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    output = Popen(cmd, stderr=STDOUT, stdout=PIPE)
    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")
    return output


def confirm_action(question, force=False):
    """confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question)
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import os
import sys
from .terminal import get_installdir, which, run_command


def generalize_spack_archive(name):
    """
    Helper function to consistently return the archive name with arch, os, no build hash
    """
    return "-".join(name.split("-")[:-1]) + ".spack"


def add_pakages_spack_repo():
    import IPython

    IPython.embed()
    import spack.repo

    spack.repo.add()


def install_spack(repo=None, branch=None):
    """
    Install spack to pakages/spack (note this is not used).
    """
    repo = repo or "https://github.com/spack/spack"
    branch = branch or "develop"
    spack_prefix = os.path.join(get_installdir(), "spack")
    return run_command(["git", "clone", "-b", branch, repo, spack_prefix])


def add_spack_to_path():
    """
    Find spack and add to path, allowing for import of spack modules
    """
    # First check for spack in environment
    spack_prefix = which("spack")
    if spack_prefix:

        # spack -> bin -> root
        spack_prefix = os.path.dirname(os.path.dirname(spack_prefix))

    else:
        # Then check for spack installed to package
        spack_prefix = os.path.join(get_installdir(), "spack")

    # Otherwise, fail
    if not os.path.exists(spack_prefix):
        sys.exit("spack must be installed! Add to path for pakages to find.")

    spack_lib_path = os.path.join(spack_prefix, "lib", "spack")
    spack_external_libs = os.path.join(spack_lib_path, "external")
    for path in [spack_lib_path, spack_external_libs]:
        sys.path.insert(0, path)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import spack.spec
from pakages.logger import logger
import spack.util.string
import pakages.repo
import six


class Spec(spack.spec.Spec):
    @property
    def package(self):
        """
        Return a custom repository that can deliver Pakages packages
        """
        if not self._package:
            self._package = pakages.repo.get(self)
        return self._package


def parse(string):
    """Returns a list of specs from an input string.
    For creating one spec, see Spec() constructor.
    """
    return


def parse_specs(packages):
    """
    Parse specs from a list of strings, and concretize
    """
    if not isinstance(packages, six.string_types):
        packages = " ".join(spack.util.string.quote(packages))

    specs = []
    for legacy in spack.spec.SpecParser().parse(packages):

        logger.info(f"Preparing spec for {legacy}")
        # Create a new Pak spec to copy (duplicate) into
        spec = Spec()
        spec._dup(legacy)

        # Always set the arch to be general
        spec.architecture = spack.spec.ArchSpec()
        spec.architecture.target = spack.target.Target("x86_64")

        spec.concretize()
        specs.append(spec)
    return specs



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient
from .helpers import parse_package_request


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)

    # Get list of packages and (optionally) repository
    packages, repo = parse_package_request(args)

    # Finally, add the repository and install packages
    if repo:
        cli.add_repository(repo.repo_dir)
    cli.install(args.packages, registry=args.registry, tag=args.tag)



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import pakages.utils as utils
from pakages.logger import logger
import requests
import os


base = "https://api.github.com"
accept = [
    "application/vnd.github.v3+json",
    "application/vnd.github.antiope-preview+json",
    "application/vnd.github.shadow-cat-preview+json",
]
accept_headers = ";".join(accept)


class GitHub:
    def __init__(self, token=None):
        self.token = token or os.environ.get("GITHUB_TOKEN")
        self.init_headers()

    def init_headers(self):
        self.headers = {"Accept": accept_headers}
        if self.token:
            self.headers["Authorization"] = "token %s" % self.token

    def clone(self, url, dest=None):
        """
        Given a URL, clone to dest. If dest not provided, create tmpdir
        """
        if not dest:
            dest = utils.get_tmpdir()

        res = utils.run_command(["git", "clone", url, dest])
        if res["return_code"] != 0:
            logger.exit(res["message"])
        return dest

    def get_org_packages(self, org):
        """
        Get organization packages.

        It doesn't seem to matter to set custom content type - these are pushed
        as containers. This is not ideal - we should be able to make an endpoint that
        is free to read with this metadata.
        """
        if not self.token:
            logger.error(
                "A token with packages:read scope is required for the Packages API."
            )
            return
        return self.get("/orgs/%s/packages" % org, {"package_type": "container"})

    def get(self, url, data=None):
        response = requests.get(base + url, headers=self.headers, params=data)
        if response.status_code != 200:
            logger.exit("Failed request to %s: %s" % (url, response.json()))



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

from pakages.client import PakClient
from pakages.logger import logger


def main(args, parser, extra, subparser):
    cli = PakClient(args.settings_file)
    installed = cli.list_installed()
    if not installed:
        logger.warning("There are no packages installed.")



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

## ContainerConfig Schema

schema_url = "https://json-schema.org/draft-07/schema/#"

## Settings.yml (loads as json)

# Currently all of these are required
settingsProperties = {
    "content_type": {"type": "string"},
    "config_editor": {"type": "string"},
    "updated_at": {"type": ["string", "null"]},
    "username": {"type": ["string", "null"]},
    "email": {"type": ["string", "null"]},
    "cache_dir": {"type": ["string", "null"]},
    # We pull from these
    "trusted_pull_registries": {"type": "array", "items": {"type": "string"}},
    # This is where we push to
    "trusted_packages_registry": {"type": "string"},
    "default_tag": {"type": "string"},
}

settings = {
    "$schema": schema_url,
    "title": "Settings Schema",
    "type": "object",
    "required": [
        "default_tag",
        "trusted_packages_registry",
        "trusted_pull_registries",
        "config_editor",
        "content_type",
    ],
    "properties": settingsProperties,
    "additionalProperties": False,
}



#!/usr/bin/env python

__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import pakages
from pakages.logger import setup_logger
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(
        description="pakages community package manager",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--settings-file",
        dest="settings_file",
        help="custom path to settings file.",
    )
    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    # Install, Uninstall and Build
    install = subparsers.add_parser(
        "install",
        description="install a Pak to the current environment",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    uninstall = subparsers.add_parser(
        "uninstall",
        description="install a Pak from the current environment",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # List installed packages
    ls = subparsers.add_parser(
        "list",
        description="list installed packages",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    push = subparsers.add_parser(
        "push",
        description="push an existing cache",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # For a push we assue we don't want default cleanup
    push.add_argument(
        "--cleanup",
        "-c",
        dest="cleanup",
        default=False,
        action="store_true",
        help="Clean up (remove) the cache after push.",
    )

    # One required, one optional arg
    push.add_argument(
        "paths",
        help="path to cache directory (optional) and GitHub packages or other oras URI (ghcr.io)",
        nargs="*",
    )

    build = subparsers.add_parser(
        "build",
        description="build into a cache",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    build.add_argument(
        "--cache-dir",
        dest="cache_dir",
        help="path to cache directory",
    )

    build.add_argument(
        "--key",
        "-k",
        dest="key",
        help="specify the gpg key hash to use",
    )
    build.add_argument(
        "--push",
        "-p",
        dest="push",
        help="push to a named oras endpoint",
    )
    build.add_argument(
        "--pushd",
        dest="push_trusted",
        action="store_true",
        default=False,
        help="push to default trusted endpoint",
    )
    build.add_argument(
        "--force",
        "-f",
        dest="force",
        action="store_true",
        default=False,
        help="force cleanup of a custom cache directory",
    )
    build.add_argument(
        "--no-cleanup",
        dest="no_cleanup",
        default=False,
        action="store_true",
        help="Given that --push is added, don't clean up the build cache.",
    )

    config = subparsers.add_parser(
        "config",
        description="update configuration settings. Use set or get to see or set information.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    config.add_argument(
        "--central",
        "-c",
        dest="central",
        help="make edits to the central config file, if a user config is default.",
        default=False,
        action="store_true",
    )

    config.add_argument(
        "params",
        nargs="*",
        help="""Set or get a config value, edit the config, add or remove a list variable, or create a user-specific config.
pakages config set key:value
pakages config get key
pakages edit
pakages config init""",
        type=str,
    )

    # Local shell with client loaded
    shell = subparsers.add_parser(
        "shell",
        description="shell into a Python session with a client.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    shell.add_argument(
        "--interpreter",
        "-i",
        dest="interpreter",
        help="python interpreter",
        choices=["ipython", "python"],
        default="ipython",
    )

    for command in [install, build, uninstall]:
        command.add_argument("packages", help="install these packages", nargs="+")

    for command in [install, build]:
        command.add_argument(
            "--registry",
            "-r",
            dest="registry",
            help="registry to use for install or build.",
        )

    for command in [install, build, push]:
        command.add_argument(
            "--tag",
            "-t",
            dest="tag",
            help="tag to use for build cache retrieval or push",
        )

    return parser


def run_main():
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """

        version = pakages.__version__

        print("\nSingularity Registry (HPC) Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(pakages.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser
                break

    if args.command == "build":
        from .build import main
    elif args.command == "config":
        from .config import main
    elif args.command == "list":
        from .ls import main
    elif args.command == "install":
        from .install import main
    elif args.command == "push":
        from .push import main
    elif args.command == "shell":
        from .shell import main
    elif args.command == "uninstall":
        from .uninstall import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run_main()



__author__ = "Vanessa Sochat, Alec Scott"
__copyright__ = "Copyright 2021-2022, Vanessa Sochat and Alec Scott"
__license__ = "Apache-2.0"

import hashlib
import errno
import os
import re
import shutil
import sys
import tempfile
import yaml

import json
from contextlib import contextmanager
from pakages.logger import logger


def read_json(filename):
    with open(filename, "r") as fd:
        content = json.loads(fd.read())
    return content


def write_json(content, filename):
    with open(filename, "w") as fd:
        fd.write(json.dumps(content, indent=4))


def read_yaml(filename):
    if not os.path.exists(filename):
        sys.exit("%s does not exist." % filename)
    with open(filename, "r") as fd:
        content = yaml.load(fd.read(), Loader=yaml.SafeLoader)
    return content


def mkdirp(dirnames):
    """
    Create one or more directories
    """
    if isinstance(dirnames, str):
        dirnames = [dirnames]
    for dirname in dirnames:
        mkdir_p(dirname)


@contextmanager
def workdir(dirname):
    """
    Provide context for a working directory, e.g.,

    with workdir(name):
       # do stuff
    """
    here = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        os.chdir(here)


def mkdir_p(path):
    """mkdir_p attempts to get the same functionality as mkdir -p
    :param path: the path to create.
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            logger.exit("Error creating path %s, exiting." % path)


def get_tmpfile(tmpdir=None, prefix=""):
    """
    Get a temporary file with an optional prefix.
    """

    # First priority for the base goes to the user requested.
    tmpdir = get_tmpdir(tmpdir)

    # If tmpdir is set, add to prefix
    if tmpdir:
        prefix = os.path.join(tmpdir, os.path.basename(prefix))

    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)

    return tmp_file


def get_tmpdir(tmpdir=None, prefix="", create=True):
    """
    Get a temporary directory for an operation.
    """
    tmpdir = tmpdir or tempfile.gettempdir()
    prefix = prefix or "pakages-tmp"
    prefix = "%s.%s" % (prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)

    if not os.path.exists(tmpdir) and create is True:
        os.mkdir(tmpdir)

    return tmpdir


def recursive_find(base, pattern=None):
    """
    Find filenames that match a particular pattern, and yield them.
    """
    # We can identify modules by finding module.lua
    for root, folders, files in os.walk(base):
        for file in files:
            fullpath = os.path.abspath(os.path.join(root, file))

            if pattern and not re.search(pattern, fullpath):
                continue

            yield fullpath


def get_file_hash(image_path, algorithm="sha256"):
    """
    Return an sha256 hash of the file based on a criteria level.
    """
    try:
        hasher = getattr(hashlib, algorithm)()
    except AttributeError:
        logger.error("%s is an invalid algorithm.")
        logger.exit(" ".join(hashlib.algorithms_guaranteed))

    with open(image_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()


def copyfile(source, destination, force=True):
    """
    Copy a file from a source to its destination.
    """
    # Case 1: It's already there, we aren't replacing it :)
    if source == destination and force is False:
        return destination

    # Case 2: It's already there, we ARE replacing it :)
    if os.path.exists(destination) and force is True:
        os.remove(destination)

    shutil.copyfile(source, destination)
    return destination


def write_file(filename, content, mode="w"):
    """
    Write content to a filename
    """
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def print_json(json_obj):
    """Print json pretty"""
    return json.dumps(json_obj, indent=4, separators=(",", ": "))


def read_file(filename, mode="r"):
    """Read a file."""
    with open(filename, mode) as filey:
        content = filey.read()
    return content
